import IPython.display as ipd
from io import BytesIO
from cairosvg import svg2png, svg2pdf, svg2eps
from .outputs import SVG

class AksharArt:
    
    def __init__(self, image, dist) -> None:
        self.image = image
        self.dist = dist
    
    def set_dim(self, width:int=None, height:int=None):
        self.image.set_dim(width, height)
        
    def aksharify(self, show=False) -> None:
        self.matrix = []
        
        for x in range(self.image.bwimg.shape[0]):
            line = []
            for y in range(self.image.bwimg.shape[1]):
                line.append(self.dist.char_dict[int(self.image.bwimg[x, y]*255)])
            self.matrix.append(line)
        
        if show:
            svg = SVG()
            svg.background_color = "#ffffff"
            svg.bold = False
            svg.fill_color = "#000000"
            art = svg.generate_art(self.matrix, self.image.image)
            ipd.display(ipd.SVG(art))
    
    def replace_char(self, char:str, x:int, y:int)-> None:
        if x<self.w and y<self.h:
            self.matrix[y][x] = char
        else:
            raise IndexError

    def replace_chars(self, chars:str, x:int, y:int) -> None:
        x, y = abs(x), abs(y)
        if x>self.w or y>self.h:
            raise IndexError
        if self.w - x >= len(chars):
            for i in range(x, x + len(chars)):
                self.replace_char(chars[i-x], i, y)
        else:
            self.replace_chars(chars[:self.w-x], x, y)
            self.replace_chars(chars[self.w-x:], 0, y+1)

    def txt_output(self, fname:str) -> None:
        text = ""
        for line_no in range(self.image.bwimg.shape[0]):
            text += "".join(self.matrix[line_no]) + "\n"
        with open(fname + ".txt", "w") as file:
            file.write(text)

    def svg_output(self, config):
        svg = config
        with open(svg.file_name + ".svg", "w", encoding="utf-8-sig") as file:
            file.write(svg.generate_art(self.matrix, self.image.image))
    
    def png_output(self, config):
        png = config
        svg = png.generate_svg(self.matrix, self.image.image)
        if png.height == None and png.width != None:
            return svg2png(bytestring=svg, write_to=png.file_name+".png", output_width=png.width)
        elif png.width == None and png.height != None:
            return svg2png(bytestring=svg, write_to=png.file_name+".png", output_height=png.height)
        elif png.height == None and png.width == None:
            return svg2png(bytestring=svg, write_to=png.file_name+".png")
        else:
            return svg2png(bytestring=svg, write_to=png.file_name+".png", output_height=png.height, output_width=png.width)

    def pdf_output(self, config):
        pdf = config
        svg = pdf.generate_svg(self.matrix, self.image.image)
        svg_io = BytesIO(svg.encode('utf-8-sig'))
        svg2pdf(file_obj=svg_io, write_to=config.file_name + ".pdf")
    
    def eps_output(self, config):
        eps = config
        svg = eps.generate_svg(self.matrix, self.image.image)
        svg_io = BytesIO(svg.encode('utf-8'))
        svg2eps(file_obj=svg_io, write_to=eps.file_name + ".eps")
    
    def html_output(self, config):
        html = config
        with open(html.file_name + ".html", "w", encoding="utf-8") as file:
            file.write(html.generate_art(self.matrix, self.image.image))




import numpy as np
import matplotlib.pyplot as plt

SORTEDCHARS = """ `.-_',~:*;!"^/\+><r=?()|7LxtcYivTljsz[]1Jnufy{}oI#FC4VX2ehk3aZw5ASbdpqUP6%9G8mKO&0EDHg$MWRNQB@"""

class Dist:
    def __init__(self, chars, unique=True, order=False) -> None:
        if unique:
            self.chars = []
            for char in chars:
                if char not in self.chars:
                    self.chars.append(char)
        else:
            self.chars = list(chars)
        if order:
            chars = []
            for char in SORTEDCHARS:
                while char in self.chars:
                    chars.append(char)
                    self.chars.remove(char)
            self.chars = chars
        self.cmap = plt.get_cmap("tab20")
        self.y = lambda x: x
        self.char_dict = {}

    def check_ele(self, list):
        if len(list) != 1 + len(self.chars):
            raise ValueError(f"{1 + len(self.chars)} number of elements expected in the list.")

    def normal_to_hex(self):
        self.check_ele(self.list)
        self.list[0], self.list[-1] = 0, 1
        self.list = 255*self.list
        self.list = self.list.astype(int)

    def hex_to_normal(self):
        self.check_ele(self.list)
        self.list = self.list/255
        self.list[0], self.list[-1] = 0, 1

    def generate_char_dict(self):
        self.check_ele(self.list)
        j = 1
        for i in range(256):
            self.char_dict[i] = self.chars[j-1]
            if not i<self.list[j]:
                j += 1
    def show(self):
        self.hex_to_normal()
        x = np.arange(0, 1.001, 0.001)
        y = self.y(x)
        plt.plot(x,y)
        # plt.fill_between(x, y, np.max(y), facecolor='red', alpha=0.5)
        colors = [self.cmap(i) for i in range(len(self.chars))] 
        for i in range(len(self.list)-1):
            a = int(self.list[i]/0.001)
            b = int(self.list[i+1]/0.001)
            np.array([y[a]])
            np.array([y[b]])
            np.array([0])
            X = np.concatenate((np.array([0]), x[a:b], np.array([0]), np.array([0])))
            Y = np.concatenate((np.array([y[a]]), y[a:b], np.array([y[b]]), np.array([y[50]])))
            plt.fill_between(X, Y, facecolor=colors[i], alpha=0.3)
            plt.fill_between(x[a:b], y[a:b], facecolor=colors[i], alpha=0.3)
        for i, color in enumerate(colors[::-1]):
            plt.plot([2], [0], marker='o', markersize=10, color=color, alpha=0.3, label=self.chars[-(i+1)]+f": {self.list[-(i+2)]:.2f}-{self.list[-(i+1)]:.2f}")
        plt.legend(fontsize=8)
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.show()
        self.normal_to_hex()


class Linear(Dist):
    def __init__(self, chars, unique=True, order=False) -> None:
        super().__init__(chars, unique, order)
        num = 255/len(self.chars)
        self.list = [0]
        for i in range(len(self.chars)):
            j = int(num * (i + 1))
            self.list.append(j)
        self.list[-1] = 255
        self.list = np.array(self.list)
        self.generate_char_dict()

class Exponential(Linear):
    def __init__(self, chars, power, unique=True, order=False) -> None:
        super().__init__(chars, unique, order)
        self.hex_to_normal()
        self.list = self.list**power
        self.normal_to_hex()
        self.y = lambda x: x**power
        self.generate_char_dict()  
        
class Normal(Linear):
    def __init__(self, chars, mean, var, unique=True, order=False) -> None:
        super().__init__(chars, unique, order)
        self.mean = mean
        self.var = var
        self.hex_to_normal()
        self.list[0], self.list[-1] = 0.0001, 0.9999
        self.list = self.normalizer(self.list)
        self.normal_to_hex()
        self.y = self.normalizer
        self.generate_char_dict()
    
    def normalizer(self, x):
        inv = (x*(1-self.mean))/(self.mean*(1-x))
        inv **= -self.var
        inv += 1
        return inv**(-1)
    
    def show(self):
        self.hex_to_normal()
        x = np.arange(0.001, 0.999, 0.001)
        y = self.y(x)
        plt.plot(x,y)
        # plt.fill_between(x, y, np.max(y), facecolor='red', alpha=0.5)
        colors = [self.cmap(i) for i in range(len(self.chars))]
        for i in range(len(self.list)-1):
            a = int(self.list[i]/0.001)
            b = int(self.list[i+1]/0.001) - 3 # Index error
            np.array([y[a]])
            np.array([y[b]])
            np.array([0])
            X = np.concatenate((np.array([0]), x[a:b], np.array([0]), np.array([0])))
            Y = np.concatenate((np.array([y[a]]), y[a:b], np.array([y[b]]), np.array([y[50]])))
            plt.title("Normal")
            plt.fill_between(X, Y, facecolor=colors[i], alpha=0.3)
            plt.fill_between(x[a:b], y[a:b], facecolor=colors[i], alpha=0.3)
        for i, color in enumerate(colors[::-1]):
            plt.plot([2], [0], marker='o', markersize=10, color=color, alpha=0.3, label=self.chars[-(i+1)]+f": {self.list[-(i+2)]:.2f}-{self.list[-(i+1)]:.2f}")
        plt.legend(fontsize=8)
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.show()
        self.normal_to_hex()


import skimage as ski
import matplotlib.pyplot as plt
import numpy as np
import requests
from io import BytesIO

class Image:

    def __init__(self, image=None, url=None) -> None:
        self.CH_CONSTANT = 1.78
        if image and url:
            raise ValueError("Provide image or link to the Image")
        elif image:
            self.obj = ski.io.imread(image)
        else:
            response = requests.get(url)
            image_data = response.content
            self.obj = ski.io.imread(BytesIO(image_data))
        self.image = self.obj
        self.bwimg = ski.color.rgb2gray(self.obj[:,:,:3])
        self.w, self.h = self.obj.shape[1], self.obj.shape[0]
        
    def show(self, bw=False):
        plt.axis('off')
        if not bw:
            plt.imshow(self.image)
        else:
            plt.imshow(self.bwimg, cmap='gray')
        plt.show()

    def set_dim(self, width:int=None, height:int=None):
        if width != None and height == None:
            w, h = width, int((self.h/self.w * width)/self.CH_CONSTANT)
        elif width == None and height != None:
            w, h = int((self.w/self.h * height) * self.CH_CONSTANT), height
        else:
            w, h = width, height
        self.image = ski.transform.resize(self.obj, (h, w), anti_aliasing=True)
        self.bwimg = ski.color.rgb2gray(self.image[:,:,:3])
    
    def normalize(self):
        self.bwimg = np.subtract(self.bwimg, np.min(self.bwimg))
        self.bwimg = np.divide(self.bwimg, np.max(self.bwimg))


LINK = "https://primepatel.github.io/aksharify/"
SVG_HEADER = '<?xml version="1.0" standalone="no"?><svg width="{}" height="{}" version="1.1" xmlns="http://www.w3.org/2000/svg" style="font-family: {}; font-size:{};"><desc>Aksharify Art</desc><rect width="100%" height="100%" fill="{}"/>'
HTML_HEADER = '<!DOCTYPE html><html lang="en"><head><meta charset="UTF-8"><meta name="viewport" content="width=device-width, initial-scale=1.0"><title>Aksharify Art</title></head><body><a href="https://primepatel.github.io/aksharify/" style="text-decoration: none">{}</a></body></html>'


class HTML:
    
    def __init__(self):
        self.title = "Aksharify Art"
        self.link = LINK # not added yet
        self.font_family = "monospace"
        self.font_size = 10
        self.bold = False
        self.file_name = "aksharify"

    def bspan(self, char:str, char_colour:tuple) -> str:
        return f"<span style='color: rgb{char_colour};'><b>{char}</b></span>"
    
    def span(self, char:str, char_colour:tuple) -> str:
        return f"<span style='color: rgb{char_colour};'>{char}</span>"

    def generate_art(self, matrix, image) -> None:
        html_content = f'<p style="font-size: {self.font_size}px; font-family: {self.font_family};">'
        w, h = image.shape[1], image.shape[0] 
        if self.bold == True:
            span = self.bspan
        else:
            span = self.span
        for line_no in range(h):
            for char_no in range(w):
                rgba = image[line_no, char_no]
                r, g, b = rgba[0]*255, rgba[1]*255, rgba[2]*255
                html_content += span(
                    f"&#{ord(matrix[line_no][char_no])};", (r,g,b)
                    )
            html_content += '<br>'
        html_content += "</p>"
        return HTML_HEADER.format(html_content)


class SVG:
    
    def __init__(self):
        self.font_family = "monospace"
        self.font_size = 10
        self.background_color = "#ffffff"
        self.link = LINK # not added yet
        self.bold = False
        self.fill_color = None
        self.file_name = "aksharify"
        self.background_color = "None"
    
    def generate_art(self, matrix, image) -> None:
        H_dis = round((self.font_size*555)/1000, 2)
        V_dis = self.font_size
        w, h = image.shape[1], image.shape[0] 
        width = int(w*H_dis)+41
        height = int(h*V_dis)+41
        file = SVG_HEADER.format(width, height, self.font_family, self.font_size, self.background_color)
        # file += f'<a href="https://primepatel.github.io/aksharify/">'
        file += '<style>text{'
        if self.bold:
            file += 'font-weight: bold;'
        if self.fill_color:
            file += 'fill:' + self.fill_color + ';'
        file += '}</style>'
        if self.fill_color:
            char_func = lambda x, char: f'<tspan x="{x}">{char}</tspan>'
            x, y = 20, 30
            for line_no in range(h):
                file += f'<text x="{x}" y="{y}">'
                for char_no in range(w):
                    file += char_func(x, f"&#{ord(matrix[line_no][char_no])};")
                    x += H_dis
                file += '</text>'
                x = 20
                y += V_dis
            # file += "</a>"
            file += "</svg>"
        else:
            rgb2hex = lambda rgba: '#{:02x}{:02x}{:02x}'.format(int(rgba[0]*255), int(rgba[1]*255), int(rgba[2]*255))
            char_func = lambda x, char, char_color: f'<tspan x="{x}" fill="{char_color}">{char}</tspan>'
            x, y = 20, 30
            for line_no in range(h):
                file += f'<text x="{x}" y="{y}">'
                for char_no in range(w):
                    file += char_func(x, f"&#{ord(matrix[line_no][char_no])};", rgb2hex(image[line_no, char_no]))
                    x += H_dis
                file += '</text>'
                x = 20
                y += V_dis
            # file += "</a>"
            file += "</svg>"
        return file


class PNG(SVG):
    def __init__(self) -> None:
        super().__init__()
        self.width = None
        self.height = None
        
    def generate_svg(self, matrix, image) -> None:
        return self.generate_art(matrix, image)


class PDF(SVG):
    def __init__(self) -> None:
        super().__init__()
        
    def generate_svg(self, matrix, image) -> None:
        return self.generate_art(matrix, image)
    

class EPS(SVG):
    def __init__(self) -> None:
        super().__init__()
        
    def generate_svg(self, matrix, image) -> None:
        return self.generate_art(matrix, image)